function [SensorX, SensorY, SensorZ] = LeverArm3DTri(CoordList, Meas, Uncert)
%Precise Lever Arm Determiation via 3D Geodetic Triangulation for Aided
%Inertial Navigation Systems

% More on Description:
% (1)   This function is written for the determination of precise lever arm
%       between an IMU and an aiding sensor such as GNSS antenna, odometer, laser scanner,
%       camera, radar altimeter, etc.

% (2)   It performs iterative weighted least squares using "lscov.m" function of MATLAB

% (3)   The inputs are:
%       - Coordinate list of reference points/markers located on the IMU, Total Station/standing point, and Aiding Sensor given in inertial sensor coordinate frame (CoordList)
%       - Measurements of slant ditance, horizontal direction, and zenit angles (Meas)
%       - Uncertinaties of distance and angular quantities (Uncert)

% (4)   The outputs are 3D coordinates of aiding sensor (X,Y,Z) in inertial sensor coordinate frame.

% (5)   Please cite as Simav and Akdoğan (2022)

%       Simav M. and Akdoğan Y.A. (2022). Accurate lever arm determination
%       via 3D triangulation for aided inertial navigation systems, Harita
%       Dergisi, 168, pp XX-YY.

% (6)   Written by Mehmet Simav (mehmet.simav@harita.gov.tr), May, 2022


% Input Formats:

% CoordList:   Cell matrix of size Nx5 (N: number of reference plus unknown points)
%               Col-1: Point Name
%               Col-2: Exact or approximate X-coordinate of Reference Points, Total Station, and Aiding Sensor in the inertial sensor coordinate frame in cm unit 
%               Col-3: Exact or approximate Y-coordinate of Reference Points, Total Station, and Aiding Sensor in the inertial sensor coordinate frame in cm unit 
%               Col-4: Exact or approximate Z-coordinate of Reference Points, Total Station, and Aiding Sensor in the inertial sensor coordinate frame in cm unit 
%               Col-5: Code for point types as described below: 
%                      1: Reference point/marker located on the IMU
%                      2: Total station/standing point
%                      3: Aiding sensor such as GNSS antenna, odometer, etc.
%                                              

%               Example: Table-1 of the reference paper (Simav and Akdoğan, 2022) 
%               CoordList = {   'L-F'      0.00      0.00	     0.00    1
%                               'L-1'     -2.73      0.25	    -1.30    1
%                               'L-2'     -2.73      0.25	    -9.30    1
%                               'L-3'     -2.73      0.25	   -17.30    1
%                               'L-4'     -4.73      0.25	    -0.50    1
%                               'L-5'     -4.73      0.25	   -18.20    1
%                               'L-6'    -14.23      0.25	    -0.50    1
%                               'L-7'    -14.23      0.25	   -18.20    1
%                               'L-8'    -23.73      0.25	    -0.50    1
%                               'L-9'    -23.73      0.25	   -18.20    1
%                               'L-10'   -33.23      0.25	    -0.50    1
%                               'L-11'   -33.23      0.25	    -1.30    1
%                               'L-12'   -35.23      0.25	    -1.30    1
%                               'L-13'   -35.23      0.25	    -9.30    1
%                               'L-14'   -35.23      0.25	   -17.30    1
%                               'L-B'    -38.00	     0.00	     0.00    1
%                               'T-1'     -9.55      4.85	   -20.20    1
%                               'T-2'    -18.55      4.85	   -20.20    1
%                               'TS'     -36.50  -1085.00     -120.50    2
%                               'SNSR'   -12.40      3.86     -140.00    3     
%                          };


% Meas  :    Cell matrix of size Mx4 (M: number of slant distance, horizontal direction, and zenith angles)
%            Col-1: Point Name (NOTE: Point names must be same as given in CoordList. Please see the example below)
%            Col-2: Slant distances between standing point (total station) and other points in meter unit 
%            Col-3: Horizontal directions between standing point (total station) and other points in decimal degree unit 
%            Col-4: Zenith angles between standing point (total station) and other points in decimal degree unit 

%            Example: Table-2 of the reference paper (Simav and Akdoğan, 2022) 
%            Meas = {'L-F'    10.851    89.3366666666667    96.3100000000000
%                    'L-1'    10.853    89.4836111111111    96.2380555555556
%                    'L-2'    10.847    89.4791666666667    95.8183333333333
%                    'L-3'    10.844    89.4744444444444    95.3966666666667
%                    'L-4'    10.852    89.5888888888889    96.2822222222222
%                    'L-5'    10.845    89.5822222222222    95.3522222222222
%                    'L-6'    10.853    90.0950000000000    96.2783333333333
%                    'L-7'    10.839    90.0863888888889    95.3513888888889
%                    'L-8'    10.851    90.5972222222222    96.2758333333333
%                    'L-9'    10.838    90.5886111111111    95.3472222222222
%                    'L-10'   10.843    91.1038888888889    96.2711111111111
%                    'L-11'   10.838    91.0961111111111    95.3436111111111
%                    'L-12'   10.843    91.2083333333333    96.2275000000000
%                    'L-13'   10.842    91.2066666666667    95.8066666666667
%                    'L-14'   10.838    91.2019444444445    95.3841666666667
%                    'L-B'    10.841    91.3627777777778    96.2997222222222
%                    'T-1'    10.873    89.8697222222222    95.2488888888889
%                    'T-2'    10.872    90.3461111111111    95.2425000000000
%                    'SNSR'   10.866    90.0230555555556    87.8950000000000
%                  };

% Uncert    :Vector of size 1x2
%            Col-1: Slant distance uncertainty in meter unit
%            Col-2: Direction (horizontal or zenith) uncertainty in decimal degree unit

%            Example: Table-3 of the reference paper (Simav and Akdoğan, 2022) 
%            Uncert = [3/1000    3/3600];

%% Let's Start

%Mirror the Z-coordinates of the points if the positive Z direction of the IMU is downward.
CoordList(:,4) = cellfun(@(x) -x,CoordList(:,4),'UniformOutput',false);

%Point Types in CoordList cell matrix
PTypes = cell2mat(CoordList(:,5));

%Total Station initial XYZ coordinates in [m]
Idx = PTypes == 2;
TSxyz = cell2mat(CoordList(Idx,2:4)) / 100;

%Sensor initial XYZ coordinates in [m]
Idx = PTypes == 3;
SensorXYZ = cell2mat(CoordList(Idx,2:4)) / 100;

%Sensor Name (Point Name for Sensor)
SensorName = CoordList{Idx,1} ;

%Number of Distance, Direction and Zenith measurements
RowNum = size(Meas,1);
n      = RowNum*3;

%Number of Unknows
u = 7;

%Measurement Type: Slant Distance, Horizontal Directions and Zenith Angles
MeasType    = [ cellstr(repmat('SlantDist',RowNum,1)) ;                             
                cellstr(repmat('HorzDirec',RowNum,1)) ; 
                cellstr(repmat('Zenith',RowNum,1))];

%Measurement points
MeasPoint = repmat(Meas(:,1),3,1);

%Degree to Radian
Deg2Rad = pi/180;

%Measurement Vector
L = [ cell2mat(Meas(:,2));                                                
      cell2mat(Meas(:,3)) * Deg2Rad; 
      cell2mat(Meas(:,4)) * Deg2Rad];
  
%Weights  
w = [ repmat(1 / Uncert(1)        , RowNum  , 1);                           
      repmat(1 / Uncert(2)*Deg2Rad, RowNum*2, 1)];
  
%Initial values of the unknowns  
x = [TSxyz' ; SensorXYZ' ; 0];

%Iteration Number
IterNum = 1;

%Iterative Solution
while 1
    %Reset the design matrix (A) and observed-computed (closure vector) (DeltaC)
    A       = zeros(n,u);
    DeltaC  = zeros(n,1);
    
    %Most recent values of Standing Point (Total Station) XYZ Coordinates
    TSCoord = x(1:3);                                                              
    for i = 1 : n
        %Measurement Point XYZ Coordinates
        if strcmp(MeasPoint{i},SensorName)
            PCoord = x(4:6)';                                                       
        else
            PCoord  = cell2mat(CoordList(ismember(CoordList(:,1),MeasPoint{i}),2:4)) / 100;
        end
        
        %Commonly used variables
        Dx = PCoord(1) - TSCoord(1);                %Xi-Xt  or  Xg-Xt
        Dy = PCoord(2) - TSCoord(2);                %Yi-Yt  or  Yg-Yt
        Dz = PCoord(3) - TSCoord(3);                %Zi-Zt  or  Zg-Zt
        
        %Slant Distance
        if strcmp(MeasType{i},'SlantDist')
            S3D    = sqrt(Dx^2+Dy^2+Dz^2);          %3D distance between Standing Point (TS) and measured point
            A(i,1) = -Dx / S3D;                     %dF/dXt
            A(i,2) = -Dy / S3D;                     %dF/dYt
            A(i,3) = -Dz / S3D;                     %dF/dZt
            if strcmp(MeasPoint{i},SensorName)
                A(i,4) = -A(i,1);                   %dF/dXg
                A(i,5) = -A(i,2);                   %dF/dYg
                A(i,6) = -A(i,3);                   %dF/dZg
            end
            DeltaC(i) = L(i) - S3D;                 %observed-computed (closure vector)
        end
        
        %Horizontal Directions
        if strcmp(MeasType{i},'HorzDirec')
            S2DSq  = Dx^2 + Dy^2;                   %2D distance squared
            A(i,1) = Dy / S2DSq;                    %dF/dXt
            A(i,2) = -Dx / S2DSq;                   %dF/dYt
            if strcmp(MeasPoint{i},SensorName)
                A(i,4) = -A(i,1);                   %dF/dXg
                A(i,5) = -A(i,2);                   %dF/dYg
            end
            A(i,7) = 1;                             %dF/dGamma
            DeltaC(i) = L(i) - atan2(Dy,Dx) - x(7); 
        end
        
        %Zenith Angles
        if strcmp(MeasType{i},'Zenith')
            S2DSq  = Dx^2 + Dy^2;                   %2D distance squared
            S2D    = sqrt(S2DSq);                   %2D distance
            Denom1 = ((S2DSq + Dz^2) / Dz) * S2D;   %Denominator for dF/dXt and %dF/dYt
            Denom2 = S2DSq + Dz^2;                  %Denominator for dF/dZt 
            A(i,1) = -Dx / Denom1;                  %dF/dXt
            A(i,2) = -Dy / Denom1;                  %dF/dYt
            A(i,3) = S2D / Denom2;                  %dF/dZt
            if strcmp(MeasPoint{i},SensorName)
                A(i,4) = -A(i,1);                   %dF/dXg
                A(i,5) = -A(i,2);                   %dF/dYg
                A(i,6) = -A(i,3);                   %dF/dZg
            end
            DeltaC(i) =  L(i) - atan2(S2D,Dz);      %observed-computed (closure vector)
        end
    end
    
    %Least Square Estimation
    [Deltax,STDx] = lscov(A,DeltaC,w);
    
    %Correction to coordinate unknowns smaller than 5 mm
    if all(Deltax(1:6) <= 5/1000)                 
        break                               
    else
        IterNum = IterNum + 1;
    end
    x = x + Deltax;
end

% Output
SensorX = x(4);
SensorY = x(5);
SensorZ = -x(6);

%Print Estimated Coordinates of Total Station and GNSS Antenna wrt to IMU L-F corner
fprintf('Number of Iteration          : %8d\n',IterNum);
fprintf('Total Station - X[m] ± STD[m]: %8.3f ± %3.2f\n',x(1),STDx(1));
fprintf('Total Station - Y[m] ± STD[m]: %8.3f ± %3.2f\n',x(2),STDx(2));
fprintf('Total Station - Z[m] ± STD[m]: %8.3f ± %3.2f\n',-x(3),STDx(3));
fprintf('Aiding Sensor - X[m] ± STD[m]: %8.3f ± %3.2f\n',x(4),STDx(4));
fprintf('Aiding Sensor - Y[m] ± STD[m]: %8.3f ± %3.2f\n',x(5),STDx(5));
fprintf('Aiding Sensor - Z[m] ± STD[m]: %8.3f ± %3.2f\n',-x(6),STDx(6));
end